@@ -61,10 +61,8 @@ extern CONST int md_short_jump_size;
 extern CONST int md_long_jump_size;
 #endif
 
-/* Used to control final evaluation of expressions that are more
-   complex than symbol + constant.  1 means set final value for simple
-   expressions, 2 means set final value for more complex expressions.  */
-int finalize_syms = 1;
+/* Used to control final evaluation of expressions.  */
+int finalize_syms = 0;
 
 int symbol_table_frozen;
 void print_fixup PARAMS ((fixS *));
@@ -760,10 +758,10 @@ adjust_reloc_syms (abfd, sec, xxx)
 	   symbols, though, since they are not in the regular symbol
 	   table.  */
 	if (sym != NULL)
-	  resolve_symbol_value (sym, finalize_syms);
+	  resolve_symbol_value (sym);
 
 	if (fixp->fx_subsy != NULL)
-	  resolve_symbol_value (fixp->fx_subsy, finalize_syms);
+	  resolve_symbol_value (fixp->fx_subsy);
 
 	/* If this symbol is equated to an undefined symbol, convert
            the fixup to being against that symbol.  */
@@ -1576,14 +1574,15 @@ write_object_file ()
       if (!changed)
 	break;
     }
+  /* Relaxation has completed.  Freeze all syms.  */
+  finalize_syms = 1;
+
   bfd_map_over_sections (stdoutput, size_seg, (char *) 0);
 #else
   relax_and_size_all_segments ();
+  finalize_syms = 1;
 #endif /* BFD_ASSEMBLER  */
 
-  /* Relaxation has completed.  Freeze all syms.  */
-  finalize_syms = 2;
-
 #if defined (BFD_ASSEMBLER) && defined (OBJ_COFF) && defined (TE_GO32)
   /* Now that the segments have their final sizes, run through the
      sections and set their vma and lma. !BFD gas sets them, and BFD gas
@@ -1902,7 +1901,7 @@ write_object_file ()
       symbolS *symp;
 
       for (symp = symbol_rootP; symp; symp = symbol_next (symp))
-	resolve_symbol_value (symp, finalize_syms);
+	resolve_symbol_value (symp);
     }
   resolve_local_symbol_values ();
 
@@ -1950,7 +1949,7 @@ write_object_file ()
 	  /* Do it again, because adjust_reloc_syms might introduce
 	     more symbols.  They'll probably only be section symbols,
 	     but they'll still need to have the values computed.  */
-	  resolve_symbol_value (symp, finalize_syms);
+	  resolve_symbol_value (symp);
 
 	  /* Skip symbols which were equated to undefined or common
              symbols.  */
@@ -2242,8 +2241,9 @@ relax_segment (segment_frag_root, segment)
 
 	    if (offset % fragP->fr_var != 0)
 	      {
-		as_bad (_("alignment padding (%lu bytes) not a multiple of %ld"),
-			(unsigned long) offset, (long) fragP->fr_var);
+		as_bad_where (fragP->fr_file, fragP->fr_line,
+			      _("alignment padding (%lu bytes) not a multiple of %ld"),
+			      (unsigned long) offset, (long) fragP->fr_var);
 		offset -= (offset % fragP->fr_var);
 	      }
 
@@ -2354,10 +2354,11 @@ relax_segment (segment_frag_root, segment)
 			    {
 			      char buf[50];
 			      sprint_value (buf, (addressT) lie->addnum);
-			      as_warn (_(".word %s-%s+%s didn't fit"),
-				       S_GET_NAME (lie->add),
-				       S_GET_NAME (lie->sub),
-				       buf);
+			      as_warn_where (fragP->fr_file, fragP->fr_line,
+					     _(".word %s-%s+%s didn't fit"),
+					     S_GET_NAME (lie->add),
+					     S_GET_NAME (lie->sub),
+					     buf);
 			    }
 			  lie->added = 1;
 			  if (fragP->fr_subtype == 0)
@@ -2452,23 +2453,31 @@ relax_segment (segment_frag_root, segment)
 		}
 
 	      case rs_space:
+		growth = 0;
 		if (symbolP)
 		  {
-		    growth = S_GET_VALUE (symbolP);
-		    if (symbol_get_frag (symbolP) != &zero_address_frag
+		    offsetT amount;
+
+		    amount = S_GET_VALUE (symbolP);
+		    if (S_GET_SEGMENT (symbolP) != absolute_section
 			|| S_IS_COMMON (symbolP)
 			|| ! S_IS_DEFINED (symbolP))
-		      as_bad_where (fragP->fr_file, fragP->fr_line,
-				    _(".space specifies non-absolute value"));
-		    fragP->fr_symbol = 0;
-		    if (growth < 0)
 		      {
-			as_warn (_(".space or .fill with negative value, ignored"));
-			growth = 0;
+			as_bad_where (fragP->fr_file, fragP->fr_line,
+				      _(".space specifies non-absolute value"));
+			/* Prevent repeat of this error message.  */
+			fragP->fr_symbol = 0;
 		      }
+		    else if (amount < 0)
+		      {
+			as_warn_where (fragP->fr_file, fragP->fr_line,
+				       _(".space or .fill with negative value, ignored"));
+			fragP->fr_symbol = 0;
+		      }
+		    else
+		      growth = (was_address + amount
+				- fragP->fr_next->fr_address);
 		  }
-		else
-		  growth = 0;
 		break;
 
 	      case rs_machine_dependent:
@@ -2488,7 +2497,7 @@ relax_segment (segment_frag_root, segment)
 		  valueT value;
 		  int size;
 
-		  value = resolve_symbol_value (fragP->fr_symbol, 0);
+		  value = resolve_symbol_value (fragP->fr_symbol);
 		  size = sizeof_leb128 (value, fragP->fr_subtype);
 		  growth = size - fragP->fr_offset;
 		  fragP->fr_offset = size;
@@ -2611,7 +2620,7 @@ fixup_segment (fixP, this_segment_type)
 
       if (sub_symbolP)
 	{
-	  resolve_symbol_value (sub_symbolP, finalize_syms);
+	  resolve_symbol_value (sub_symbolP);
 	  if (add_symbolP == NULL || add_symbol_segment == absolute_section)
 	    {
 	      if (add_symbolP != NULL)
